from regex_enumerator import RegexEnumerator
from .test_function import f_finite, f_infinite

def test_single_alternative():
    regexEnumerator = RegexEnumerator('a|b')
    possibilities = ['a', 'b']

    f_finite(regexEnumerator, possibilities)

def test_single_alternative_with_quantifier():
    regexEnumerator = RegexEnumerator('a|b*')
    possibilities = ['a', '', 'b', 'bb', 'bbb', 'bbbb', 'bbbbb']

    f_infinite(regexEnumerator, possibilities)

def test_single_alternative_with_quantifier_plus():
    regexEnumerator = RegexEnumerator('a+|b')
    possibilities = ['b', 'a', 'aa', 'aaa', 'aaaa', 'aaaaa']

    f_infinite(regexEnumerator, possibilities)

def test_multiple_alternatives():
    regexEnumerator = RegexEnumerator('a|b|c')
    possibilities = ['a', 'b', 'c']

    f_finite(regexEnumerator, possibilities)

def test_alternative_with_char_and_class():
    regexEnumerator = RegexEnumerator('a|[b-d]')
    possibilities = ['a', 'b', 'c', 'd']

    f_finite(regexEnumerator, possibilities)

def test_alternative_with_class_with_quantifier_0_and_char():
    regexEnumerator = RegexEnumerator('[a-c]{0}|d')
    possibilities = ['', 'd']

    f_finite(regexEnumerator, possibilities)


from regex_enumerator import RegexEnumerator


def f_finite(regexEnumerator: RegexEnumerator, possibilities: list[str]):
    while len(possibilities) != 0:
        res = regexEnumerator.next()
        assert res in possibilities
        possibilities.remove(res)
        
    assert regexEnumerator.done

def f_infinite(regexEnumerator: RegexEnumerator, possibilities: list[str]):
    while len(possibilities) != 0:
        res = regexEnumerator.next()
        assert res in possibilities
        possibilities.remove(res)
    
    assert not regexEnumerator.done


from .regex_parser import RegexParser
from .regex_tree import RegexTree


class RegexEnumerator:
    def __init__(self, regex: str):
        self.regex: str = regex
        parser = RegexParser()
        self.regexTree: RegexTree = parser.parse(regex)
        self.current: list[str] = list(self.regexTree.current)
        self.done: bool = self.regexTree.done and len(self.current) == 0

    def next(self) -> str | None:

        if len(self.current) != 0:
            res = self.current.pop()
            self.done = self.regexTree.done and len(self.current) == 0
            return res

        while True:
            if self.regexTree.done:
                self.done = True
                return None
            self.current = list(self.regexTree.next())
            if len(self.current) != 0:
                break

        res = self.current.pop()
        self.done = self.regexTree.done and len(self.current) == 0
        return res



from regex_enumerator import RegexEnumerator
from .test_function import f_finite, f_infinite

def test_single_char_class():
    regexEnumerator = RegexEnumerator('[a]')
    possibilities = ['a']

    f_finite(regexEnumerator, possibilities)

def test_single_char_class_with_quantifier():
    regexEnumerator = RegexEnumerator('[a]*')
    possibilities = ['', 'a', 'aa', 'aaa', 'aaaa', 'aaaaa']

    f_infinite(regexEnumerator, possibilities)

def test_multiple_char_class():
    regexEnumerator = RegexEnumerator('[a-c]')
    possibilities = ['a', 'b', 'c']

    f_finite(regexEnumerator, possibilities)

def test_multiple_char_class_with_quantifier():
    regexEnumerator = RegexEnumerator('[a-c]{1,2}')
    possibilities = ['a', 'b', 'c', 'aa', 'ab', 'ac', 'ba', 'bb', 'bc', 'ca', 'cb', 'cc']

    f_finite(regexEnumerator, possibilities)

def test_multiple_char_class_with_quantifier_0():
    regexEnumerator = RegexEnumerator('[a-c]{0}')
    possibilities = ['']

    f_finite(regexEnumerator, possibilities)

def test_multiple_char_class_with_quantifier_plus():
    regexEnumerator = RegexEnumerator('[a-b]+')
    possibilities = ['a', 'b', 'aa', 'ab', 'ba', 'bb', 'aaa', 'aab', 'aba', 'abb', 'baa', 'bab', 'bba', 'bbb']

    f_infinite(regexEnumerator, possibilities)

def test_2_multiple_char_class_with_quantifier_question():
    regexEnumerator = RegexEnumerator('[a-cf-g]?')
    possibilities = ['', 'a', 'b', 'c', 'f', 'g']

    f_finite(regexEnumerator, possibilities)


from regex_enumerator import RegexEnumerator
from .test_function import f_finite, f_infinite

def test_char_class_surronded_by_other_chars():
    regexEnumerator = RegexEnumerator('a[0-9]b')
    possibilities = ['a0b', 'a1b', 'a2b', 'a3b', 'a4b', 'a5b', 'a6b', 'a7b', 'a8b', 'a9b']

    f_finite(regexEnumerator, possibilities)


from .regex_tree import Alternative, CharClasses, RegexTree


class RegexParser:
    charset = [chr(c) for c in range(32, 127)]

    def parse(self, regex) -> RegexTree:
        _, regexTree = self._parseRegex(regex, False)
        return regexTree

    def _parseRegex(self, regex: str, to_close: bool) -> tuple[int, RegexTree]:
        alternatives: list[Alternative] = []
        charClassesList: list[CharClasses | RegexTree] = []
        min_len = 1
        max_len = 1
        closed = not to_close

        i = 0
        while i < len(regex):
            if regex[i] == '(':
                incr, subTree = self._parseRegex(regex[i + 1:], True)
                charClassesList.append(subTree)
                i += incr + 1
            elif regex[i] == ')':
                if to_close:
                    i += 1
                    incr, min_len, max_len = self._parseQuantifier(regex[i:])
                    i += incr
                    closed = True
                    break
                raise ValueError('Invalid regex')
            elif regex[i] == '|':
                alternatives.append(Alternative(charClassesList))
                charClassesList = []
                i += 1
            else:
                incr, charClasses = self._parseCharClasses(regex[i:])
                charClassesList.append(charClasses)
                i += incr

        if not closed:
            raise ValueError('Invalid regex')

        alternatives.append(Alternative(charClassesList))
        return i, RegexTree(alternatives, min_len, max_len)

    def _parseCharClasses(self, regex: str) -> tuple[int, CharClasses]:
        i = 0
        chars_list: list[str] = []

        if regex[i] == '[':
            i += 1
            inc, chars_list = self._parseCharClass(regex[i:])
            i += inc
        elif regex[i] == '\\':
            raise NotImplementedError(
                'Escape characters are not supported yet')
        elif regex[i] == '.':
            chars_list = [chr(c) for c in range(0, 128)]
        else:
            chars_list = [regex[i]]
            i = 1

        incr, min_len, max_len = self._parseQuantifier(regex[i:])
        i += incr

        charClasses = CharClasses(chars_list, min_len, max_len)
        return i, charClasses

    def _parseCharClass(self, regex: str) -> tuple[int, list[str]]:
        i = 0
        chars_list: list[str] = []
        first_char = None
        range_divider = False
        negated = False

        if regex[i] == '^':
            negated = True
            i += 1

        while i < len(regex) and regex[i] != ']':
            if regex[i] == '\\':
                raise NotImplementedError(
                    'Escape characters are not supported yet')
            elif first_char is None:
                first_char = regex[i]
            elif regex[i] == '-':
                range_divider = True
            elif range_divider:
                chars_list.extend([chr(c) for c in range(
                    ord(first_char), ord(regex[i]) + 1)])
                first_char = None
                range_divider = False
            else:
                if first_char is not None:
                    chars_list.append(first_char)
                    first_char = None
                chars_list.append(regex[i])
            i += 1

        if regex[i] != ']':
            raise ValueError('Invalid character class')

        i += 1

        if range_divider:
            chars_list.append('-')
        if first_char is not None:
            chars_list.append(first_char)

        if negated:
            chars_list = [c for c in self.charset if c not in chars_list]

        return i, chars_list

    def _parseQuantifier(self, regex: str) -> tuple[int, int, int]:
        i = 0

        if len(regex) == 0:
            return 0, 1, 1

        match regex[i]:
            case '*':
                return 1, 0, None
            case '+':
                return 1, 1, None
            case '?':
                return 1, 0, 1
            case '{':
                i += 1
                incr, min_len, max_len = self._parseMinMax(regex[i:])
                return i + incr, min_len, max_len
            case _:
                return 0, 1, 1

    def _parseMinMax(self, regex: str) -> tuple[int, int, int]:
        i = 0
        min_len = 0
        max_len = 0

        while i < len(regex) and regex[i] == ' ':
            i += 1

        if i >= len(regex) or regex[i] not in '0123456789':
            raise ValueError('Invalid quantifier')

        while i < len(regex) and regex[i] in '0123456789':
            min_len = min_len * 10 + int(regex[i])
            i += 1

        while i < len(regex) and regex[i] == ' ':
            i += 1

        if i >= len(regex):
            raise ValueError('Invalid quantifier')
        elif regex[i] == '}':
            return i + 1, min_len, min_len
        elif regex[i] != ',':
            raise ValueError('Invalid quantifier')

        i += 1

        while i < len(regex) and regex[i] == ' ':
            i += 1

        if i >= len(regex) or regex[i] not in '0123456789}':
            raise ValueError('Invalid quantifier')

        if regex[i] == '}':
            return i + 1, min_len, None

        while i < len(regex) and regex[i] in '0123456789':
            max_len = max_len * 10 + int(regex[i])
            i += 1

        if max_len < min_len:
            raise ValueError('Invalid quantifier')

        while i < len(regex) and regex[i] == ' ':
            i += 1

        if i >= len(regex) or regex[i] != '}':
            raise ValueError('Invalid quantifier')

        i += 1

        return i, min_len, max_len



class RegexTree:
    pass


class CharClasses:
    def __init__(self, chars_list: list[str], min_len: int, max_len: int):
        self.index = 0
        self.chars: str = ''.join(sorted(set(''.join(chars_list))))
        self.min_len = min_len
        self.max_len = max_len
        self.done = False
        self.base = len(self.chars)

        if self.base == 0 or self.max_len == 0:
            self.done = True
            self.current: set[str] = {''}
            return

        if self.max_len is None:
            self.max_index = None
        elif self.base == 1:
            self.max_index = self.max_len - self.min_len
        else:
            self.max_index = ((self.base ** self.min_len - self.base **
                              (self.max_len + 1)) // (1 - self.base)) - 1

        self.current: set[str] = {self.calculate()}

    def calculate(self) -> str:
        if self.max_len is not None and self.index >= self.max_index:
            self.done = True

        if self.base == 1:
            return self.chars[0] * (self.min_len + self.index)

        res = []

        num = self.base ** self.min_len + self.index

        while num > 1:
            res.append(self.chars[num % self.base])
            num //= self.base

        return ''.join(reversed(res))

    def next(self) -> str:
        assert not self.done

        self.index += 1
        res = self.calculate()
        self.current.add(res)
        return res


class Alternative:
    def __init__(self, classes_or_regex_list: list[CharClasses | RegexTree]):
        self.index = 0
        self.classes_or_regex_list: list[CharClasses |
                                         RegexTree] = classes_or_regex_list

        i = 0
        while i < len(self.classes_or_regex_list):
            if self.classes_or_regex_list[i].done and len(self.classes_or_regex_list[i].current) == 0:
                self.classes_or_regex_list.pop(i)
            else:
                i += 1

        self.base = len(self.classes_or_regex_list)

        if self.base == 0:
            self.done = True
            self.current: set[str] = {''}
            return

        self.done = False
        self.current: set[str] = self.calculate()

    def next(self) -> set[str]:
        assert not self.done

        res = set()
        self.index += 1
        for i in range(self.index, self.index + self.base):
            if self.classes_or_regex_list[i % self.base].done:
                self.index += 1
            else:
                break

        if 0 == self.index % self.base:
            if isinstance(self.classes_or_regex_list[0], CharClasses):
                res.add(self.classes_or_regex_list[0].next())
            else:
                res.update(self.classes_or_regex_list[0].next())
        else:
            res.update(self.classes_or_regex_list[0].current)

        done = self.classes_or_regex_list[0].done

        for i, classes_or_regex in enumerate(self.classes_or_regex_list[1:], start=1):
            temp = set()

            if i == self.index % self.base:
                next_str = classes_or_regex.next()
                if isinstance(next_str, str):
                    for prv_str in res:
                        temp.add(prv_str + next_str)
                else:
                    for prv_str in res:
                        for string in next_str:
                            temp.add(prv_str + string)
            else:
                for prv_str in res:
                    for string in classes_or_regex.current:
                        temp.add(prv_str + string)

            done = done and classes_or_regex.done
            res = temp

        self.done = done

        new_res = res - self.current
        self.current.update(new_res)
        return new_res

    def calculate(self) -> set[str]:
        res = set()
        res.update(self.classes_or_regex_list[0].current)

        done = self.classes_or_regex_list[0].done

        for classes_or_regex in self.classes_or_regex_list[1:]:
            temp = set()
            done = done and classes_or_regex.done
            for prv_str in res:
                for string in classes_or_regex.current:
                    temp.add(prv_str + string)

            res = temp

        self.done = done
        return res


class RegexTree:
    def __init__(self, alternatives: list[Alternative], min_len: int, max_len: int):
        self.alternatives: list[Alternative] = alternatives
        self.min_len = min_len
        self.max_len = max_len

        i = 0
        while i < len(self.alternatives):
            if self.alternatives[i].done and len(self.alternatives[i].current) == 0:
                self.alternatives.pop(i)
            else:
                i += 1

        self.base = len(self.alternatives)

        if self.base == 0 or self.max_len == 0:
            self.done = True
            self.current: set[str] = set()
            return

        self.done = False
        self.gen_charset = False
        self.index_charset = 0
        self.index_repetition = 0
        self.done_repetition = False
        self.current_charset: set[str] = self.calculate_charset()
        self.current: set[str] = self.calculate()

    def next(self) -> set[str]:
        assert not self.done

        if self.done_charset:
            self.gen_charset = False
        elif self.done_repetition:
            self.gen_charset = True

        if self.gen_charset:
            next_charset: set[str] = self.next_charset()
            # optimize it by using only the new charset
        else:
            if not self.done_repetition:
                self.index_repetition += 1

        res: set[str] = self.calculate()

        self.gen_charset = not self.gen_charset
        new_res = res - self.current
        self.current.update(new_res)
        return new_res

    def calculate(self) -> set[str]:
        if self.max_len is not None and self.index_repetition + self.min_len >= self.max_len:
            self.done_repetition = True
            if self.done_charset:
                self.done = True

        if self.index_repetition + self.min_len == 0:
            return {''}

        res = set(self.current_charset)

        for i in range(1, self.min_len + self.index_repetition):
            temp = set()
            for prv_str in res:
                for string in self.current_charset:
                    temp.add(prv_str + string)
            res = temp

        return res

    def next_charset(self) -> set[str]:
        assert not self.done_charset

        self.index_charset += 1

        for i in range(self.index_charset, self.index_charset + self.base):
            if self.alternatives[i % self.base].done:
                self.index_charset += 1
            else:
                break

        res = self.alternatives[self.index_charset % self.base].next()

        done_charset = True

        for alternative in self.alternatives:
            if not alternative.done:
                done_charset = False
                break

        new_res = res - self.current_charset
        self.current_charset.update(new_res)
        self.done_charset = done_charset
        return new_res

    def calculate_charset(self) -> set[str]:
        res = set()

        done_charset = True

        for alternative in self.alternatives:
            res.update(alternative.current)
            done_charset = done_charset and alternative.done

        self.done_charset = done_charset

        return res



from regex_enumerator import RegexEnumerator
from .test_function import f_finite, f_infinite

def test_empty_regex():
    regexEnumerator = RegexEnumerator('')
    possibilities = ['']

    f_finite(regexEnumerator, possibilities)

def test_single_char():
    regexEnumerator = RegexEnumerator('a')
    possibilities = ['a']

    f_finite(regexEnumerator, possibilities)

def test_single_char_with_quantifier():
    regexEnumerator = RegexEnumerator('a*')
    possibilities = ['', 'a', 'aa', 'aaa', 'aaaa', 'aaaaa']

    f_infinite(regexEnumerator, possibilities)

def test_single_char_with_quantifier_plus():
    regexEnumerator = RegexEnumerator('a+')
    possibilities = ['a', 'aa', 'aaa', 'aaaa', 'aaaaa']

    f_infinite(regexEnumerator, possibilities)

def test_single_char_with_quantifier_question():
    regexEnumerator = RegexEnumerator('a?')
    possibilities = ['', 'a']

    f_finite(regexEnumerator, possibilities)

def test_single_char_with_quantifier_braces():
    regexEnumerator = RegexEnumerator('a{2}')
    possibilities = ['aa']

    f_finite(regexEnumerator, possibilities)

def test_single_char_with_quantifier_braces_comma():
    regexEnumerator = RegexEnumerator('a{2,}')
    possibilities = ['aa', 'aaa', 'aaaa', 'aaaaa']

    f_infinite(regexEnumerator, possibilities)

def test_single_char_with_quantifier_braces_comma_max():
    regexEnumerator = RegexEnumerator('a{2,4}')
    possibilities = ['aa', 'aaa', 'aaaa']

    f_finite(regexEnumerator, possibilities)

def test_single_char_with_quantifier_0():
    regexEnumerator = RegexEnumerator('a{0}')
    possibilities = ['']

    f_finite(regexEnumerator, possibilities)


from regex_enumerator import RegexEnumerator
from .test_function import f_finite, f_infinite

def test_single_char():
    regexEnumerator = RegexEnumerator('(a)')
    possibilities = ['a']

    f_finite(regexEnumerator, possibilities)

def test_single_char_class():
    regexEnumerator = RegexEnumerator('([a])')
    possibilities = ['a']

    f_finite(regexEnumerator, possibilities)

def test_multiple_char_class():
    regexEnumerator = RegexEnumerator('([a-c])')
    possibilities = ['a', 'b', 'c']

    f_finite(regexEnumerator, possibilities)

def test_single_char_with_quantifier():
    regexEnumerator = RegexEnumerator('(a)*')
    possibilities = ['', 'a', 'aa', 'aaa', 'aaaa', 'aaaaa']

    f_infinite(regexEnumerator, possibilities)


from setuptools import setup, find_packages

setup(
    name='regex_enumerator',
    version='0.1.0',
    packages=find_packages(include=['regex_enumerator', 'regex_enumerator.*']),
    description='Enumerate all strings that match a given regex',
    author='Vincenzo Greco',
    author_email='grecovincenzo98@gmail.com',
    extras_require={
        'dev': ['pytest', 'pytest-benchmark'],  # Development dependencies, e.g., for testing
    },
)